// MichaelisMentenReaction.cpp: implementation of the MichaelisMentenReaction class.
//
//////////////////////////////////////////////////////////////////////

#include "MichaelisMentenReaction.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

MichaelisMentenReaction::MichaelisMentenReaction(Chemical *Enzyme, Chemical *Substrate, Chemical *Product, RateConstant *k_2, RateConstant *K_m, std::string reactionName)
:Reaction(reactionName)
{
	// stoichiometry
	this->Enzyme = chemicals.size();
	chemicals.push_back(Enzyme);
	numberOfEachChemicalChangedByReaction.push_back(0);

	this->Substrate = chemicals.size();
	chemicals.push_back(Substrate);
	numberOfEachChemicalChangedByReaction.push_back(-1);

	this->Product = chemicals.size();
	chemicals.push_back(Product);
	numberOfEachChemicalChangedByReaction.push_back(+1);

	// rates
	this->k_2 = rateConstants.size();
	rateConstants.push_back(k_2);

	this->K_m = rateConstants.size();
	rateConstants.push_back(K_m);

	// put enough dummy chemicalJacobian elements on the vector 
	// - they are recomputed every time you ask for the chemicalJacobian
	chemicalJacobian.push_back(new JElement(-1,-1));
	chemicalJacobian.push_back(new JElement(-1,-1));

	this->SetTeXForm();
}

MichaelisMentenReaction::~MichaelisMentenReaction()
{

}

double MichaelisMentenReaction::GetRate() const
{
	return rateConstants[k_2]->GetValue()*chemicals[Enzyme]->GetAmount()*
		(chemicals[Substrate]->GetAmount()/(chemicals[Substrate]->GetAmount() + rateConstants[K_m]->GetValue()));
}

void MichaelisMentenReaction::SetTeXForm()
{
	strstream stream;
	// TeX form of the reaction
	stream << rateConstants[k_2]->GetTeXName().c_str();
	stream << " \\left[" << chemicals[Enzyme]->GetTeXName().c_str() << "\\right]";
	stream << " {\\left[" << chemicals[Substrate]->GetTeXName().c_str() << "\\right]";
	stream << "\\over {";
	stream << "\\left[" << chemicals[Substrate]->GetTeXName().c_str() << "\\right]";
	stream << " + " << rateConstants[K_m]->GetTeXName().c_str() << "}}";
	// add a NULL
	stream << ends;
	// set the TeX string equal to what's in the stream
	m_sTeXForm = stream.str();
	// allow memory to be freed
	stream.rdbuf()->freeze(0);
}

std::vector<Reaction::JElement *> *MichaelisMentenReaction::GetChemicalJacobian()
{
	chemicalJacobian[0]->SetWithRespectTo(chemicals[Enzyme]->GetChemicalNumber());
	chemicalJacobian[0]->SetJValue((rateConstants[k_2]->GetValue()*chemicals[Substrate]->GetAmount())/(chemicals[Substrate]->GetAmount() + rateConstants[K_m]->GetValue()));
	
	chemicalJacobian[1]->SetWithRespectTo(chemicals[Substrate]->GetChemicalNumber());
	chemicalJacobian[1]->SetJValue((rateConstants[k_2]->GetValue()*rateConstants[K_m]->GetValue()*chemicals[Enzyme]->GetAmount())
		/ ((chemicals[Substrate]->GetAmount() + rateConstants[K_m]->GetValue())*(chemicals[Substrate]->GetAmount() + rateConstants[K_m]->GetValue())) );

	return &chemicalJacobian;
}
