// LaubLoomisDictyNetworkMinimizableDirector.cpp: implementation of the CLaubLoomisDictyNetworkMinimizableDirector class.
//
//////////////////////////////////////////////////////////////////////

#include "LaubLoomisDictyNetworkMinimizableDirector.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CLaubLoomisDictyNetworkMinimizableDirector::CLaubLoomisDictyNetworkMinimizableDirector()
{
	// create networks,experiments,observers,movers,minimizables
	double maxSimTime = 91.0;
	double stepSize = 0.0001;
	double frequency = 1.0;
	int nCells = 1;
	double moveTol = 1.0e-06;

	m_dGammaSquared = 1.0e-07;
	
	// flags for all minimizables
	bool rateFlag = true;
	bool chemFlag = false;
	bool logsFlag = false;
	double timeWeight = 1.0;
	double rateWeight = 0.0;
	double chemWeight = 0.0;

	int cnt = 0;

	// expt. 1 - Roos 1977
	// networks and data
	_networkList.push_back(new CLaubLoomisDictyNetwork());
	_experimentList.push_back(new Experiment(_networkList[cnt],"Roos et. al. 1977"));
	// put data files here
	_experimentList[cnt]->AddDataFileToList("../DataFiles/TimeSeries/cAMPiRoos.dat");
	// mover and plotter
	_moverList.push_back(new CQualityControlCashKarpMover(frequency,stepSize,moveTol));
	_plotterList.push_back(new GnuPlotterTimeSeriesObserver());
	// title
	_plotterList[cnt]->SetTitle("D. discoideum cAMP Production:Roos et. al. 1977:ODE");
	// minimizable
	_minimizableList.push_back(new SingleNetworkMinimizable());
	// parameters
	_minimizableList[cnt]->Initialize(_experimentList[cnt],_moverList[cnt],nCells,rateFlag,chemFlag,logsFlag,timeWeight,rateWeight,chemWeight);
	_minimizableList[cnt]->SetRateConstantsWeight(0.0);
	// attach observer
	_minimizableList[cnt]->Attach(_plotterList[cnt]);
	cnt++;
	
	// expt. 2 - Aubry 1997
	// networks and data
	_networkList.push_back(new CLaubLoomisDictyNetwork);
	_experimentList.push_back(new Experiment(_networkList[cnt],"Aubry et. al. 1997"));
	// put data files here
	_experimentList[cnt]->AddDataFileToList("../DataFiles/TimeSeries/dictyERK2Aubry.dat");
	// mover and plotter
	_moverList.push_back(new CQualityControlCashKarpMover(frequency,stepSize,moveTol));
	_plotterList.push_back(new GnuPlotterTimeSeriesObserver());
	// title
	_plotterList[cnt]->SetTitle("WT D. discoideum ERK2 Activation:Aubry et. al. 1997:ODE");
	// minimizable
	_minimizableList.push_back(new SingleNetworkMinimizable());
	// parameters
	_minimizableList[cnt]->Initialize(_experimentList[cnt],_moverList[cnt],nCells,rateFlag,chemFlag,logsFlag,timeWeight,rateWeight,chemWeight);
	_minimizableList[cnt]->SetRateConstantsWeight(0.0);
	// attach observer
	_minimizableList[cnt]->Attach(_plotterList[cnt]);
	cnt++;

	// expt. 3 - Aubry 1997
	// networks and data
	_networkList.push_back(new CLaubLoomisDictyNetwork);
	_experimentList.push_back(new Experiment(_networkList[cnt],"Aubry et. al. 1997"));
	// put data files here
	_experimentList[cnt]->AddDataFileToList("../DataFiles/TimeSeries/dictyERK2PKANullAubry.dat");
	// mover and plotter
	_moverList.push_back(new CQualityControlCashKarpMover(frequency,stepSize,moveTol));
	_plotterList.push_back(new GnuPlotterTimeSeriesObserver());
	// title
	_plotterList[cnt]->SetTitle("PKA- D. discoideum ERK2 Activation:Aubry et. al. 1997:ODE");
	// minimizable
	_minimizableList.push_back(new SingleNetworkMinimizable());
	// parameters
	_minimizableList[cnt]->Initialize(_experimentList[cnt],_moverList[cnt],nCells,rateFlag,chemFlag,logsFlag,timeWeight,rateWeight,chemWeight);
	_minimizableList[cnt]->SetRateConstantsWeight(0.0);
	// attach observer
	_minimizableList[cnt]->Attach(_plotterList[cnt]);
	cnt++;
	
	// allocate the master residuals list
	int masterResidualsSize = 0;
	for(int i = 0; i < _minimizableList.size(); i++)
	{
		int subRes = _minimizableList[i]->GetNResiduals();
		masterResidualsSize += subRes;
		cout << subRes << " residuals in minimizable " << i << endl;
	}
	if(m_dGammaSquared > 0.0)
	{
		masterResidualsSize += 1;
	}
	Allocate(masterResidualsSize);

	cout << "Master Residuals size : " << masterResidualsSize << endl;

	// define the experiments
	DefineExperiments();

}

CLaubLoomisDictyNetworkMinimizableDirector::~CLaubLoomisDictyNetworkMinimizableDirector()
{
	for(int i = 0; i < _plotterList.size(); i++)
	{
		_minimizableList[i]->Detach(_plotterList[i]);
		delete _plotterList[i];
	}

}

void CLaubLoomisDictyNetworkMinimizableDirector::DefineExperiments()
{
	
	// treatment with 30 uM cAMP
	_minimizableList[1]->GetExperiment()->GetReactionNetwork()->GetChemical(5)->SetInitialAmount(300000*30);
	_minimizableList[1]->GetExperiment()->GetReactionNetwork()->ChemicalReset();
	
	// treatment with 30 uM cAMP, PKA Null
	_minimizableList[2]->GetExperiment()->GetReactionNetwork()->GetChemical(5)->SetInitialAmount(300000*30);
	_minimizableList[2]->GetExperiment()->GetReactionNetwork()->ChemicalReset();
	_minimizableList[2]->GetExperiment()->GetReactionNetwork()->RemoveReaction(5);
	_minimizableList[2]->GetExperiment()->GetReactionNetwork()->RemoveReaction(3);
	_minimizableList[2]->GetExperiment()->GetReactionNetwork()->RemoveReaction(2);
	_minimizableList[2]->GetExperiment()->GetReactionNetwork()->RemoveReaction(1);
	
}

double CLaubLoomisDictyNetworkMinimizableDirector::GetParameter(int parIndex)
{
	// for this director, everyone has the same parameters so any set 
	// can be returned
	return _minimizableList[0]->GetParameter(parIndex);
}

int CLaubLoomisDictyNetworkMinimizableDirector::GetNParameters()
{
	// ditto as for GetParameters()
	return _minimizableList[0]->GetNParameters();
}
