// StochasticSensitivityAnalysis.cpp: implementation of the CStochasticSensitivityAnalysis class.
//
//////////////////////////////////////////////////////////////////////

#include "StochasticSensitivityAnalysis.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CStochasticSensitivityAnalysis::CStochasticSensitivityAnalysis()
{

}

CStochasticSensitivityAnalysis::CStochasticSensitivityAnalysis(bool readFlag, bool hessianFlag, int nEnsembleSize, int seed, CParameterFilter *pFilter, double funcAccuracy, double temperature)
{
	m_bSVDFlag = false;
	m_bReadFlag = readFlag;
	m_bHessianFlag = hessianFlag;
	m_iRNGSeed = seed;
	m_iNEnsembleSize = nEnsembleSize;
	m_dTemperature = temperature;
	m_dRescale = 1.0/5.0;
	m_dAcceptanceRatio = 0.0;
	m_dRWValue = 0.0;
	m_dMNormSquared = 0.0;
	m_dFuncAccuracy = funcAccuracy;
	m_pFilter = pFilter;
	m_pRNG = new Rand(seed);
}

CStochasticSensitivityAnalysis::~CStochasticSensitivityAnalysis()
{
	delete m_pRNG;
	delete [] m_pdSingularValues;
	for(int i = 0; i < m_vpEnsemble.size(); i++)
	{
		delete m_vpEnsemble[i];
	}
}

void CStochasticSensitivityAnalysis::DoAnalysis(double *pFittedParameters, Minimizable *pMinimizable)
{
	// cost difference between trial move and current move
	double deltaCost = 0.0;
	double quadratic = 0.0;
	double newCost = 0.0;
	// set the number of parameters
	m_iNParameters = pMinimizable->GetNParameters();

	// downcast to a sum-of-squares minimizable (this routine isn't designed to handle
	// anything else)
	NLLSMinimizable *pNLLS = dynamic_cast<NLLSMinimizable *>(pMinimizable);
	if(0 == pNLLS) {throw std::runtime_error("Minimizable is not a sum of squares!");}

	m_iNResiduals = pNLLS->GetNResiduals();

	// allocate memory
	m_pdHessian = new double *[m_iNParameters];
	m_pdSamplingMatrix = new double *[m_iNParameters];
	m_pdMatrixV = new double *[m_iNParameters];
	m_pdHessian[0] = new double[m_iNParameters*m_iNParameters];
	m_pdSamplingMatrix[0] = new double[m_iNParameters*m_iNParameters];
	m_pdMatrixV[0] = new double[m_iNParameters*m_iNParameters];
	for(int nPar = 1; nPar < m_iNParameters; nPar++)
	{
		m_pdHessian[nPar] = &(m_pdHessian[0][nPar*m_iNParameters]);
		m_pdSamplingMatrix[nPar] = &(m_pdSamplingMatrix[0][nPar*m_iNParameters]);
		m_pdMatrixV[nPar] = &(m_pdMatrixV[0][nPar*m_iNParameters]);
	}
	m_pdJacobian = new double *[m_iNResiduals];
	m_pdJacobian[0] = new double[m_iNResiduals*m_iNParameters];
	for(int nR = 1; nR < m_iNResiduals; nR++)
	{
		m_pdJacobian[nR] = &(m_pdJacobian[0][nR*m_iNParameters]);
	}

	// allocate space for current and trial parameters
	double *pParCurrent = new double[m_iNParameters];
	double *pParTrial = new double[m_iNParameters];
	double *pParDelta = new double[m_iNParameters];
	m_pdSingularValues = new double[m_iNParameters];

	// open a file to write the ratio of the actual change in cost to that 
	// predicted by the Hessian at the fitted values
	fstream *pfRatio = new fstream("ratio.par",ios::out);
	*pfRatio << "##################################################################" << endl;
	*pfRatio << "# delta(C)/(1/2) k.H0.k" << endl;
	*pfRatio << "# TEMP = " << m_dTemperature << endl;
	*pfRatio << "##################################################################" << endl;

	// set the current parameters equal to the best parameters
	for(int i = 0; i < m_iNParameters; i++)
	{
		pParCurrent[i] = pFittedParameters[i];
	}
	// get the cost at the minimum; if you're using a synchronized integrator, this
	// has the additional benefit of setting up the list of stepsizes before the 
	// derivative is computed
	// OLD
	// double currentCost = pNLLS->ObjectiveFunction(pFittedParameters);
	// NEW
	double currentCost = pNLLS->F(pFittedParameters,m_dTemperature);
	// add the minimum to the ensemble
	AddParametersToEnsemble(pFittedParameters,pNLLS->GetResiduals(),0.0,currentCost);
	
	cout << "Cost at minimum is " << currentCost << endl;

	if(m_bReadFlag)
	{
		ReadHessian();
	}
	else
	{
		// compute the Hessian at the minimum
		if(m_bHessianFlag)
		{
			ComputeTrueHessian(pFittedParameters,pNLLS);
		}
		else
		{
			ComputeApproximateHessian(pFittedParameters,pNLLS);
		}
	}

	// now sample away
	double currentEnsembleSize = 1;
	int nTrials = 0;
	int nAccept = 0;
	while(currentEnsembleSize < m_iNEnsembleSize)
	{
		// sample
		quadratic = SampleGaussian(pParDelta);
		//quadratic = SampleWithoutExplicitHessian(pParDelta);

		// construct the trial move
		ConstructTrialMove(pParTrial,pParCurrent,pParDelta);
		
		// OLD
		// evaluate the cost at the new parameter values
		//newCost = pNLLS->ObjectiveFunction(pParTrial);
		//deltaCost = newCost - currentCost;

		// NEW
		newCost = pNLLS->F(pParTrial,m_dTemperature);
		deltaCost = newCost - currentCost;

		// write the cost ratio
		*pfRatio << deltaCost/quadratic << endl;

		// if we don't accept, just repeat
		if(AcceptMove(quadratic,deltaCost))
		{
			nAccept++;
			m_dAcceptanceRatio += 1.0;
			cout << "Move accepted . . ." << endl;
			// copy parameters
			for(int i = 0; i < m_iNParameters; i++)
			{
				pParCurrent[i] = pParTrial[i];
				currentCost = newCost;
			}
			currentEnsembleSize++;
			// increment the RW value
			m_dRWValue += m_dRescale*m_dMNormSquared;
			// add another member to the ensemble
			AddParametersToEnsemble(pParCurrent,pNLLS->GetResiduals(),quadratic,currentCost);
			// notify observers
			Notify();
		}
	
		nTrials++;
	}

	cout << "Acceptance Ratio : " << ((double) nAccept)/((double) nTrials) << endl;

	// temporary arrays
	double *minEigen = new double[m_iNParameters];
	double *eigenP = new double[m_iNParameters];

	// before returning, write costs/parameters/etc. to a file
	fstream *pfPars = new fstream("ensemble.par",ios::out);
	fstream *pfQuad = new fstream("quadratic.par",ios::out);
	fstream *pfCost = new fstream("cost.par",ios::out);
	fstream *pfResid = new fstream("residuals.par",ios::out);
	fstream *pfEigP = new fstream("eigenpars.par",ios::out);
	*pfPars << "##################################################################" << endl;
	*pfQuad << "##################################################################" << endl;
	*pfCost << "##################################################################" << endl;
	*pfResid << "##################################################################" << endl;
	*pfEigP << "##################################################################" << endl;
	*pfPars << "# PARAMETER ENSEMBLE - POUND IS SEPARATOR" << endl;
	*pfPars << "# TEMP = " << m_dTemperature << endl;
	*pfQuad << "# VALUE OF (1/2)k.H.k FOR ALL MEMBERS OF ENSEMBLE" << endl;
	*pfQuad << "# TEMP = " << m_dTemperature << endl;
	*pfCost << "# ACTUAL COST (NOT DELTA) FOR ALL MEMBERS OF ENSEMBLE" << endl;
	*pfCost << "# TEMP = " << m_dTemperature << endl;
	*pfResid << "# RESIDUALS FOR ALL MEMBERS OF ENSEMBLE - POUND IS SEPARATOR" << endl;
	*pfResid << "# TEMP = " << m_dTemperature << endl;
	*pfEigP << "# EIGENPARAMETER ENSEMBLE - DOUBLE NEWLINE IS SEPARATOR" << endl;
	*pfEigP << "# TEMP = " << m_dTemperature << endl;
	*pfResid << "##################################################################" << endl;
	*pfPars << "##################################################################" << endl;
	*pfQuad << "##################################################################" << endl;
	*pfCost << "##################################################################" << endl;
	*pfEigP << "##################################################################" << endl;
	for(int ensNum = 0; ensNum < m_vpEnsemble.size(); ensNum++)
	{
		*pfQuad << setprecision(10) << m_vpEnsemble[ensNum]->GetQuadratic() << endl;
		*pfCost << setprecision(10) << m_vpEnsemble[ensNum]->GetCost() << endl;
		for(int j = 0; j < m_vpEnsemble[ensNum]->GetParameterSize(); j++)
		{
			*pfPars << m_vpEnsemble[ensNum]->GetParameter(j) << endl;
		}
		for(j = 0; j < m_vpEnsemble[ensNum]->GetResidualsSize(); j++)
		{
			*pfResid << m_vpEnsemble[ensNum]->GetResidual(j) << endl;
		}
		// single pound symbol for the parameter file and residuals file
		*pfPars << "#" << endl; 
		*pfResid << "#" << endl;
	}
	
	// now do the eigenparameters
	// first the minimum (0th member of the ensemble)
	for(int m = 0; m < m_iNParameters; m++)
	{
		minEigen[m] = 0.0;
	}
	// compute eigenparameters
	for(int j = 0; j < m_iNParameters; j++)
	{
		double parameter = log(m_vpEnsemble[0]->GetParameter(j));
		for(int k = 0; k < m_iNParameters; k++)
		{
			minEigen[k] += parameter*(m_pdMatrixV[j][k]);
		}
	}
	// now print it (first line is just zeroes
	for(j = 0; j < m_iNParameters; j++)
	{
		*pfEigP << 0.0 << endl;
	}
	*pfEigP << endl << endl;

	
	for(i = 1; i < m_vpEnsemble.size(); i++)
	{
		// zero the eigenparameters
		for(int m = 0; m < m_iNParameters; m++)
		{
			eigenP[m] = 0.0;
		}
		// compute eigenparameters
		for(int j = 0; j < m_iNParameters; j++)
		{
			double parameter = log(m_vpEnsemble[i]->GetParameter(j));
			for(int k = 0; k < m_iNParameters; k++)
			{
				eigenP[k] += parameter*(m_pdMatrixV[j][k]);
			}
		}
		// now print the eigenparameters
		for(j = 0; j < m_iNParameters; j++)
		{
			*pfEigP << eigenP[j] - minEigen[j] << endl;
		}
		*pfEigP << endl << endl;
	}	
	
	delete pfPars;
	delete pfQuad;
	delete pfCost;
	delete pfRatio;
	delete pfResid;

	delete [] minEigen;
	delete [] eigenP;

	// free memory
	delete [] pParCurrent;
	delete [] pParTrial;
	delete [] pParDelta;
	delete [] m_pdHessian[0];
	delete [] m_pdHessian;
	delete [] m_pdSamplingMatrix[0];
	delete [] m_pdSamplingMatrix;
	delete [] m_pdMatrixV[0];
	delete [] m_pdMatrixV;
	delete [] m_pdJacobian[0];
	delete [] m_pdJacobian;

	return;
}

void CStochasticSensitivityAnalysis::ComputeApproximateHessian(double *pParameters, NLLSMinimizable *pNLLS)
{
	double transparameter = 0.0;
	double saveparameter = 0.0;
	double parplus = 0.0;
	double parminus = 0.0;
	double *fxPlusPlus = new double[m_iNResiduals];
	double *fxPlus = new double[m_iNResiduals];
	double *fx = new double[m_iNResiduals];
	double *fxMinus = new double[m_iNResiduals];
	double *fxMinusMinus = new double[m_iNResiduals];

	// compute residuals/cost at f(x)
	double chiSq = pNLLS->ObjectiveFunction(pParameters);
	for(int k = 0; k < m_iNResiduals; k++)
	{
		fx[k] = (pNLLS->GetResiduals())[k];
	}

	// USE FOR CENTERED DIFFERENCE (5 pt as well?)
	double stepscale = pow(m_dFuncAccuracy,0.3333333);
	
	// 3 PT CENTERED DIFFERENCE
	// loop over parameters
	for(int i = 0; i < m_iNParameters; i++)
	{
		// transform the parameter and save its old value
		saveparameter = pParameters[i];
		transparameter = m_pFilter->Operator(pParameters[i]);

		double h = stepscale;
		if(fabs(transparameter) > DBL_EPSILON*1000)
		{
			h = stepscale*fabs(transparameter);
		}

		parplus = transparameter + h;
		parminus = transparameter - h;
			
		// forward step
		pParameters[i] = m_pFilter->OperatorInverse(transparameter + h);
		double chiSqPlus = pNLLS->ObjectiveFunction(pParameters);
		for(int k = 0; k < m_iNResiduals; k++)
		{
			fxPlus[k] = (pNLLS->GetResiduals())[k];
		}

		// backward step
		pParameters[i] = m_pFilter->OperatorInverse(transparameter - h);
		double chiSqMinus = pNLLS->ObjectiveFunction(pParameters);
		for(k = 0; k < m_iNResiduals; k++)
		{
			fxMinus[k] = (pNLLS->GetResiduals())[k];
		}
		
		// reset old parameter
		pParameters[i] = saveparameter;

		// fill in jacobian
		for(int j = 0; j < m_iNResiduals; j++)
		{
			double delta = (fxPlus[j] - fxMinus[j])/(2*h);
			m_pdJacobian[j][i] = delta;
		}
	}
	
	// compute (1/2) of the scaled hessian (it is symmetric), and fill in 
	// the other half
	for(i = 0; i < m_iNParameters; i++)
	{
		for(int k = i; k < m_iNParameters; k++)
		{
			m_pdHessian[i][k] = 0.0;
			for(int j = 0; j < m_iNResiduals; j++)
			{
				m_pdHessian[i][k] += m_pdJacobian[j][i]*m_pdJacobian[j][k];
			}
			m_pdHessian[k][i] = m_pdHessian[i][k];
		}
	}

	delete [] fxPlus;
	delete [] fx;
	delete [] fxMinus;

	cout << "Hessian calculated . . ." << endl;
	// write the approximate Hessian to a file for later use
	fstream *pfHessian = new fstream("approxH.par",ios::out);
	*pfHessian << "##################################################################" << endl;
	*pfHessian << "# APPROXIMATE LM HESSIAN AT MINIMUM" << endl;
	*pfHessian << "##################################################################" << endl;
	for(int row = 0; row < m_iNParameters; row++)
	{
		for(int col = 0; col < m_iNParameters; col++)
		{
			*pfHessian << setprecision(10) << m_pdHessian[row][col] << endl;
		}
	}
	delete pfHessian;

	cout << "Approximate Hessian written to file \"approxH.par\"" << endl;

	// also write the jacobian to a file for later use
	fstream *pfJacobian = new fstream("lmjacobian.par",ios::out); 
	*pfJacobian << "##################################################################" << endl;
	*pfJacobian << "# LEVENBERG-MARQUARDT JACOBIAN AT MINIMUM" << endl;
	*pfJacobian << "##################################################################" << endl;
	for(row = 0; row < m_iNResiduals; row++)
	{
		for(int col = 0; col < m_iNParameters; col++)
		{
			*pfJacobian << setprecision(10) << m_pdJacobian[row][col] << endl;
		}
	}
	delete pfJacobian;

	// Hessian has been recalculated, so set the SVD flag to true
	m_bSVDFlag = true;

	return;
}


void CStochasticSensitivityAnalysis::ComputeTrueHessian(double *pParameters, NLLSMinimizable *pNLLS)
{
	double transpi = 0.0;
	double transpj = 0.0;
	double savepi = 0.0;
	double savepj = 0.0;
	double chiSq,chiSqp,chiSqm;
	double chiSqpp,chiSqmm,chiSqpm,chiSqmp;

	// compute residuals/cost at f(x)
	chiSq = pNLLS->ObjectiveFunction(pParameters);

	// DON'T FOOL WITH THIS UNLESS YOU CHANGE DIFFERENCING FORMULA
	double stepscale = pow(m_dFuncAccuracy,0.3333333);

	// compute all (nParameters*(nParameters + 1))/2 unique hessian elements
	for(int i = 0; i < m_iNParameters; i++)
	{
		for(int j = i; j < m_iNParameters; j++)
		{
			// transform parameters i and j and save their old values
			savepi = pParameters[i];
			savepj = pParameters[j];
			transpi = m_pFilter->Operator(pParameters[i]);
			transpj = m_pFilter->Operator(pParameters[j]);

			// stepsizes - these are optimal for these formulas
			double stepscalei = stepscale;
			double stepscalej = stepscale;
			double hi = stepscalei;
			double hj = stepscalej;
			if(fabs(transpi) > DBL_EPSILON)
			{
				hi = stepscalei*fabs(transpi);
			}
			if(fabs(transpj) > DBL_EPSILON)
			{
				hj = stepscalej*fabs(transpj);
			}

			// lowest order formulas for mixed second derivative

			if(i == j)
			{
				pParameters[i] = m_pFilter->OperatorInverse(transpi + hi);
				chiSqp = pNLLS->ObjectiveFunction(pParameters);
				pParameters[i] = savepi;
				pParameters[i] = m_pFilter->OperatorInverse(transpi - hi);
				chiSqm = pNLLS->ObjectiveFunction(pParameters);
				pParameters[i] = savepi;

				m_pdHessian[i][i] = chiSqp - 2*chiSq + chiSqm;
				m_pdHessian[i][i] = 0.5*m_pdHessian[i][i]/(hi*hi);
			}
			else
			{			
				// f(xi + hi, xj + h)
				pParameters[i] = m_pFilter->OperatorInverse(transpi + hi);
				pParameters[j] = m_pFilter->OperatorInverse(transpj + hj);
				chiSqpp = pNLLS->ObjectiveFunction(pParameters);
				pParameters[i] = savepi;
				pParameters[j] = savepj;

				// f(xi + hi, xj - hj)
				pParameters[i] = m_pFilter->OperatorInverse(transpi + hi);
				pParameters[j] = m_pFilter->OperatorInverse(transpj - hj);
				chiSqpm = pNLLS->ObjectiveFunction(pParameters);
				pParameters[i] = savepi;
				pParameters[j] = savepj;

				// f(xi - hi, xj + hj)
				pParameters[i] = m_pFilter->OperatorInverse(transpi - hi);
				pParameters[j] = m_pFilter->OperatorInverse(transpj + hj);
				chiSqmp = pNLLS->ObjectiveFunction(pParameters);
				pParameters[i] = savepi;
				pParameters[j] = savepj;
	
				// f(xi - hi, xj - hj)
				pParameters[i] = m_pFilter->OperatorInverse(transpi - hi);
				pParameters[j] = m_pFilter->OperatorInverse(transpj - hj);
				chiSqmm = pNLLS->ObjectiveFunction(pParameters);
				pParameters[i] = savepi;
				pParameters[j] = savepj;


				// compute hessian element and fill in its symmetric equivalent
				m_pdHessian[i][j] = chiSqpp - chiSqpm - chiSqmp + chiSqmm;
				m_pdHessian[i][j] = 0.5*m_pdHessian[i][j]/(4*hi*hj);
				m_pdHessian[j][i] = m_pdHessian[i][j];
			}

		} // end j loop
	} // end i loop
	
	cout << "True hessian calculated . . ." << endl;
	
	// write the approximate Hessian to a file for later use
	fstream *pfHessian = new fstream("trueH.par",ios::out);
	*pfHessian << "##################################################################" << endl;
	*pfHessian << "# TRUE HESSIAN AT MINIMUM" << endl;
	*pfHessian << "##################################################################" << endl;
	for(int row = 0; row < m_iNParameters; row++)
	{
		for(int col = 0; col < m_iNParameters; col++)
		{
			*pfHessian << setprecision(10) << m_pdHessian[row][col] << endl;
		}
	}
	delete pfHessian;

	cout << "True Hessian written to file \"trueH.par\"" << endl;
	// Hessian has been recalculated, so set the SVD flag to true
	m_bSVDFlag = true;

	return;
}

void CStochasticSensitivityAnalysis::ReadHessian()
{
	char linebuf[1000];
	// open the file H.par for reading
	ifstream *pfHessian = new ifstream("H.par");
	// the file has three comment lines (begin with '#'), so read them in 
	// and throw them out
	for(int i = 0; i < 3; i++)
	{
		pfHessian->getline(linebuf,sizeof(linebuf));
	}
	// now read in the NParameters*NParameters elements
	for(int row = 0; row < m_iNParameters; row++)
	{
		for(int col = 0; col < m_iNParameters; col++)
		{
			pfHessian->getline(linebuf,sizeof(linebuf));
			istrstream *dataStream = new istrstream(linebuf);
			*dataStream >> m_pdHessian[row][col];
		}
	}
	delete pfHessian;
	cout << "Hessian read in from file \"H.par\"" << endl;
	// set SVD flag to true so Hessian is decomposed in sampling routine
	m_bSVDFlag = true;
	return;
}

double CStochasticSensitivityAnalysis::SampleWithoutExplicitHessian(double *pParDelta)
{
	double quadratic = 0.0;
	double *pRandVec = new double[m_iNParameters];

	// do the decomposition if the flag tells us to do so (basically once during the
	// entire analysis)
	if(m_bSVDFlag)
	{
		cout << "Decomposing Jacobian to find matrix square root . . ." << endl;
		
		// decompose the hessian
		char JOBU = 'A';
		char JOBVT = 'A';
		int M = m_iNResiduals;
		int N = m_iNParameters;
		int minMN = __min(M,N);
		int maxMN = __max(M,N);

		// matrix to decompose (J) in column-major order
		double *A = new double[m_iNParameters*m_iNResiduals];
		int counter = 0;
		for(int j = 0; j < m_iNParameters; j++)
		{
			for(int i = 0; i < m_iNResiduals; i++)
			{
				A[counter] = m_pdJacobian[i][j];
				counter++;
			}
		}
		
		double *D = new double[minMN];
		double *E = new double[minMN];
		double *TAUQ = new double[minMN];
		double *TAUP = new double[minMN];

		double *PT = new double[maxMN*maxMN];
		double *Q = new double[maxMN*maxMN];
		double *ATEMP = new double[M*N];

		int LDA = M;
		int LDU = M;
		int LDVT = M;
		int LWORK = 5*maxMN;
		double *WORK = new double[LWORK];
		int INFO = 0;

		// the input/return arguments to the mkl SVD functions are different depending
		// upon whether the problem is over- or underdetermined

		if( M >= N )
		{
			cout << "Using decomposition procedure for M >= N . . ." << endl;
			// put the matrix into band diagonal form
			DGEBRD(&M,&N,A,&LDA,D,E,TAUQ,TAUP,WORK,&LWORK,&INFO);
			// XXX debug
			if(INFO != 0)
			{
				cout << "ERROR : INFO = " << INFO << " in DGEBRD" << endl;
			}
			// now obtain the N leading columns of Q and copy them into our array
			for(int l = 0; l < M*N; l++)
			{
				ATEMP[l] = A[l];
			}
			char VECT = 'Q';
			int K = N;
			DORGBR(&VECT,&M,&N,&K,ATEMP,&LDA,TAUQ,WORK,&LWORK,&INFO);
			// XXX debug
			if(INFO != 0)
			{
				cout << "ERROR : INFO = " << INFO << " in DORGBR (U)" << endl;
			}
			for(l = 0; l < M*N; l++)
			{
				Q[l] = ATEMP[l];
			}
		
			// now obtain PT
			for(l = 0; l < M*N; l++)
			{
				ATEMP[l] = A[l];
			}
			VECT = 'P';
			K  = M;
			DORGBR(&VECT,&N,&N,&K,ATEMP,&LDA,TAUP,WORK,&LWORK,&INFO);
			if(INFO != 0)
			{
				cout << "ERROR : INFO = " << INFO << " in DORGBR (VT)" << endl;
			}
			for(l = 0; l < N*N; l++)
			{
				PT[l] = ATEMP[l];
			}

			// obtain the SVD decomposition of A, using the matrices found above
			char UPLO = 'U';
			int NCVT = N;
			int NRU = M;
			int NCC = 0;
			double *C = 0;
			int LDC = 1;
			LDVT = N;
			LDU = M;
			DBDSQR(&UPLO,&N,&NCVT,&NRU,&NCC,D,E,PT,&LDVT,Q,&LDU,C,&LDC,WORK,&INFO);
			if(INFO != 0)
			{
				cout << "ERROR : INFO = " << INFO << " in DBDSQR (VT)" << endl;
			}

			cout << "Matrix decomposed. Condition number is " << D[0]/D[N-1] << endl;
						
			// scroll through the singular values and find the ones whose inverses will be 
			// huge and set them to zero; also, load up the array of singular values that 
			// we store.  M >= N, so the number of singular values returned equals the
			// number of parameters
			double cutoff = D[0]*1.0e-06;
			for(int i = 0; i < N; i++)
			{
				m_pdSingularValues[i] = D[i];
				if(D[i] < cutoff) {D[i] = 0;}
				else {D[i] = 1.0/D[i];}
			}
			// D[i] now equals 1/D[i]

			// now fill in the sampling matrix ("square root" of the Hessian)
			for(i = 0; i < m_iNParameters; i++)
			{
				for(j = 0; j < m_iNParameters; j++)
				{
					m_pdSamplingMatrix[i][j] = PT[i*m_iNParameters + j];
					m_pdMatrixV[i][j] = PT[i*m_iNParameters + j];
				}
			}			
			
			// open up a file to which we dump the eigenvectors (cols. of V)
			fstream *pfEigs = new fstream("matrixV.par",ios::out);
			*pfEigs << "##################################################################" << endl;
			*pfEigs << "# EIGENVECTORS OF H - \"#\" IS SEPARATOR" << endl;
			*pfEigs << "##################################################################" << endl;
			for(i = 0; i < m_iNParameters; i++)
			{
				for(j = 0; j < m_iNParameters; j++)
				{
					*pfEigs << setprecision(10) << m_pdSamplingMatrix[j][i] << endl;
				}
				*pfEigs << "#" << endl;
			}
			delete pfEigs;

			for(i = 0; i < m_iNParameters; i++)
			{
				for(j = 0; j < m_iNParameters; j++)
				{
					m_pdSamplingMatrix[i][j] = m_pdSamplingMatrix[i][j]*D[j];
				}
			}
		
		}
		else
		{
			cout << "Using decomposition procedure for M < N . . ." << endl;
			// put the matrix into band diagonal form
			DGEBRD(&M,&N,A,&LDA,D,E,TAUQ,TAUP,WORK,&LWORK,&INFO);
			// XXX debug
			if(INFO != 0)
			{
				cout << "ERROR : INFO = " << INFO << " in DGEBRD" << endl;
			}
			// now obtain the whole M X M matrix Q and copy it into our array
			for(int l = 0; l < M*N; l++)
			{
				ATEMP[l] = A[l];
			}
			char VECT = 'Q';
			int K = N;
			DORGBR(&VECT,&M,&M,&K,ATEMP,&LDA,TAUQ,WORK,&LWORK,&INFO);
			// XXX debug
			if(INFO != 0)
			{
				cout << "ERROR : INFO = " << INFO << " in DORGBR (U)" << endl;
			}
			for(l = 0; l < M*M; l++)
			{
				Q[l] = ATEMP[l];
			}
		
			// now obtain the m leading rows of PT
			for(l = 0; l < M*N; l++)
			{
				ATEMP[l] = A[l];
			}
			VECT = 'P';
			K  = M;
			DORGBR(&VECT,&M,&N,&K,ATEMP,&LDA,TAUP,WORK,&LWORK,&INFO);
			if(INFO != 0)
			{
				cout << "ERROR : INFO = " << INFO << " in DORGBR (VT)" << endl;
			}
			for(l = 0; l < M*N; l++)
			{
				PT[l] = ATEMP[l];
			}

			// obtain the SVD decomposition of A, using the matrices found above
			char UPLO = 'L';
			int NCVT = N;
			int NRU = M;
			int NCC = 0;
			double *C = 0;
			int LDC = 1;
			LDU = M;
			LDVT = M;
			DBDSQR(&UPLO,&M,&NCVT,&NRU,&NCC,D,E,PT,&LDVT,Q,&LDU,C,&LDC,WORK,&INFO);
			if(INFO != 0)
			{
				cout << "ERROR : INFO = " << INFO << " in DBDSQR" << endl;
			}

			// in this case (M < N) there is are N - M guaranteed 0 singular values, so
			// we have to remember that when computing matrix products and such
			cout << "Matrix decomposed. Submatrix condition number is " << D[0]/D[M-1] << endl;
			// scroll through the singular values and find the ones whose inverses will be 
			// huge and set them to zero; also, load up the array of singular values that 
			// we store.  the second loop fills in the SVs that are exactly zero			
			double cutoff = D[0]*1.0e-06;
			for(int i = 0; i < M; i++)
			{
				m_pdSingularValues[i] = D[i];
				if(D[i] < cutoff) {D[i] = 0;}
				else {D[i] = 1.0/D[i];}
			}
			for(i = M; i < N; i++)
			{
				m_pdSingularValues[i] = 0.0;
			}
			// D[i] now equals 1/D[i], except for the zeroes at the end
			
			// now fill in the sampling matrix ("square root" of the Hessian); the matrix
			// will have N - M zero columns, corresponding to the N - M zero rows of 
			// PT
			for(i = 0; i < m_iNParameters; i++)
			{
				for(j = 0; j < m_iNResiduals; j++)
				{
					m_pdSamplingMatrix[i][j] = PT[i*m_iNResiduals + j];
					m_pdMatrixV[i][j] = PT[i*m_iNResiduals + j];
				}
				for(j = m_iNResiduals; j < m_iNParameters; j++)
				{
					m_pdSamplingMatrix[i][j] = 0.0;
					m_pdMatrixV[i][j] = 0.0;
				}
			}

			// open up a file to which we dump the eigenvectors (cols. of V)
			fstream *pfEigs = new fstream("matrixV.par",ios::out);
			*pfEigs << "##################################################################" << endl;
			*pfEigs << "# EIGENVECTORS OF H - \"#\" IS SEPARATOR" << endl;
			*pfEigs << "##################################################################" << endl;
			for(i = 0; i < m_iNParameters; i++)
			{
				for(j = 0; j < m_iNParameters; j++)
				{
					*pfEigs << setprecision(10) << m_pdSamplingMatrix[j][i] << endl;
				}
				*pfEigs << "#" << endl;
			}
			delete pfEigs;

			for(i = 0; i < m_iNParameters; i++)
			{
				for(j = 0; j < m_iNResiduals; j++)
				{
					m_pdSamplingMatrix[i][j] = m_pdSamplingMatrix[i][j]*D[j];
				}
			}
		}
		
		delete [] A;
		delete [] ATEMP;
		delete [] TAUQ;
		delete [] TAUP;
		delete [] Q;
		delete [] PT;
		delete [] D;
		delete [] E;
		delete [] WORK;

		
		// don't compute the matrix square root every time
		m_bSVDFlag = false;
		// dump the singular values for the matrix to a file
		fstream *svd = new fstream("singval.par",ios::out);
		*svd << "##################################################################" << endl;
		*svd << "# SINGULAR VALUES OF J AT FITTED PARAMETERS" << endl;
		*svd << "##################################################################" << endl;
		for(int i = 0; i < m_iNParameters; i++)
		{
			*svd << setprecision(10) << m_pdSingularValues[i] << endl;
		}
		delete svd;
		
	}
	// sample the gaussian in either case

	// draw random numbers r ~ N(0,1)
	for(int i = 0; i < m_iNParameters; i++)
	{
		pRandVec[i] = m_pRNG->gaussian(1.0);
	}
	
	// now rescale by the sampling matrix, including the appropriate temperature
	// factor
	// RIGHT MULTIPLICATION
	for(i = 0; i < m_iNParameters; i++)
	{
		pParDelta[i] = 0.0;
		for(int j = 0; j < m_iNParameters; j++)
		{
			pParDelta[i] += sqrt(m_dRescale)*m_pdSamplingMatrix[i][j]*pRandVec[j];
		}
	}
		
	// calculate and return the value of the quadratic form, simply (1/2)xT.JT.J.x
	// = (1/2) (Jx)T (Jx)
	double *tempVec = new double[m_iNResiduals];	

	for(i = 0; i < m_iNResiduals; i++)
	{
		tempVec[i] = 0.0;
		for(int j = 0; j < m_iNParameters; j++)
		{
			tempVec[i] += m_pdJacobian[i][j]*pParDelta[j];
		}
		quadratic += 0.5*tempVec[i]*tempVec[i];
	}
	
	delete [] tempVec;
	delete [] pRandVec;

	return quadratic;
}

double CStochasticSensitivityAnalysis::SampleGaussian(double *pParDelta)
{
	double quadratic = 0.0;
	double *pRandVec = new double[m_iNParameters];

	// do the decomposition if the flag tells us to do so (basically once during the
	// entire analysis)
	if(m_bSVDFlag)
	{
		cout << "Decomposing Hessian to find matrix square root . . ." << endl;
		
		// decompose the hessian
		char JOBU = 'A';
		char JOBVT = 'A';
		int M = m_iNParameters;
		int N = m_iNParameters;
	
		// matrix A in column-major order
		double *A = new double[m_iNParameters*m_iNParameters];
		int counter = 0;
		for(int j = 0; j < m_iNParameters; j++)
		{
			pParDelta[j] = 0.0;
			for(int i = 0; i < m_iNParameters; i++)
			{
				A[counter] = m_pdHessian[i][j];
				counter++;
			}
		}
	
		double *D = new double[m_iNParameters];
		double *E = new double[m_iNParameters];
		double *TAUQ = new double[m_iNParameters];
		double *TAUP = new double[m_iNParameters];

		double *PT = new double[m_iNParameters*m_iNParameters];
		double *Q = new double[m_iNParameters*m_iNParameters];
		double *ATEMP = new double[m_iNParameters*m_iNParameters];

		int LDA = m_iNParameters;
		int LDU = m_iNParameters;
		int LDVT = m_iNParameters;
		int LWORK = 5*M;
		double *WORK = new double[LWORK];
		int INFO = 0;

		// put the matrix into band diagonal form
		DGEBRD(&M,&N,A,&LDA,D,E,TAUQ,TAUP,WORK,&LWORK,&INFO);	
		// now obtain Q and copy it into our array
		for(int l = 0; l < m_iNParameters*m_iNParameters; l++)
		{
			ATEMP[l] = A[l];
		}
		char VECT = 'Q';
		int K = M;
		DORGBR(&VECT,&M,&N,&K,ATEMP,&LDA,TAUQ,WORK,&LWORK,&INFO);
		for(l = 0; l < m_iNParameters*m_iNParameters; l++)
		{
			Q[l] = ATEMP[l];
		}
		// now obtain PT
		for(l = 0; l < m_iNParameters*m_iNParameters; l++)
		{
			ATEMP[l] = A[l];
		}
		VECT = 'P';
		DORGBR(&VECT,&M,&N,&K,ATEMP,&LDA,TAUP,WORK,&LWORK,&INFO);
		for(l = 0; l < m_iNParameters*m_iNParameters; l++)
		{
			PT[l] = ATEMP[l];
		}

		// obtain the SVD decomposition of A, using the matrices found above
		char UPLO = 'U';
		int NCVT = m_iNParameters;
		int NRU = m_iNParameters;
		int NCC = 0;
		double *C = 0;
		int LDC = 1;
		LDVT = m_iNParameters;
		LDU = m_iNParameters;
		DBDSQR(&UPLO,&N,&NCVT,&NRU,&NCC,D,E,PT,&LDVT,Q,&LDU,C,&LDC,WORK,&INFO);
		cout << "Matrix decomposed. Condition number is " << D[0]/D[m_iNParameters-1] << endl;

		// scroll through the singular values and find the ones whose inverses will be 
		// huge and set them to zero; also, load up the array of singular values that 
		// we store
		//double cutoff = D[0]*1.0e-06;
		double cutoff = 1.0;
		for(int i = 0; i < m_iNParameters; i++)
		{
			m_pdSingularValues[i] = D[i];
			if(1.0/D[i] > cutoff) {D[i] = cutoff;}
			else {D[i] = 1.0/D[i];}
		}
		// D[i] now equals 1/D[i]

		// now fill in the sampling matrix ("square root" of the Hessian)
		for(i = 0; i < m_iNParameters; i++)
		{
			for(j = 0; j < m_iNParameters; j++)
			{
				m_pdSamplingMatrix[i][j] = PT[i*m_iNParameters + j];
				m_pdMatrixV[i][j] = PT[i*m_iNParameters + j];
			}
		}

		// open up a file to which we dump the eigenvectors (cols. of V)
		fstream *pfEigs = new fstream("matrixV.par",ios::out);
		*pfEigs << "##################################################################" << endl;
		*pfEigs << "# EIGENVECTORS OF H - \"#\" IS SEPARATOR" << endl;
		*pfEigs << "##################################################################" << endl;
		for(i = 0; i < m_iNParameters; i++)
		{
			for(j = 0; j < m_iNParameters; j++)
			{
				*pfEigs << setprecision(10) << m_pdSamplingMatrix[j][i] << endl;
			}
			*pfEigs << "#" << endl;
		}
		delete pfEigs;
		
		for(i = 0; i < m_iNParameters; i++)
		{
			for(j = 0; j < m_iNParameters; j++)
			{
				m_pdSamplingMatrix[i][j] = m_pdSamplingMatrix[i][j]*sqrt(D[j]);
			}
		}

		// Square of the norm of the last row of the sampling matrix
		for(i = 0; i < m_iNParameters; i++)
		{
			m_dMNormSquared += m_pdSamplingMatrix[m_iNParameters-1][i]*m_pdSamplingMatrix[m_iNParameters-1][i];
		}
		
		delete [] A;
		delete [] ATEMP;
		delete [] TAUQ;
		delete [] TAUP;
		delete [] Q;
		delete [] PT;
		delete [] D;
		delete [] E;
		delete [] WORK;

		// don't compute the matrix square root every time
		m_bSVDFlag = false;
		// dump the singular values for the matrix to a file
		fstream *svd = new fstream("singval.par",ios::out);
		*svd << "##################################################################" << endl;
		*svd << "# SINGULAR VALUES OF H AT FITTED PARAMETERS" << endl;
		*svd << "##################################################################" << endl;
		for(i = 0; i < m_iNParameters; i++)
		{
			*svd << setprecision(10) << m_pdSingularValues[i] << endl;
		}
		delete svd;
	}
	// sample the gaussian in either case

	// draw random numbers r ~ N(0,1)
	for(int i = 0; i < m_iNParameters; i++)
	{
		pRandVec[i] = m_pRNG->gaussian(1.0);
	}
	
	// now rescale by the sampling matrix, including the appropriate
	// reweighting by the control parameter.
	
	// RIGHT MULTIPLICATION
	for(i = 0; i < m_iNParameters; i++)
	{
		pParDelta[i] = 0.0;
		for(int j = 0; j < m_iNParameters; j++)
		{
			pParDelta[i] += sqrt(m_dRescale)*m_pdSamplingMatrix[i][j]*pRandVec[j];
		}
	}
	
	
	// calculate and return the value of the quadratic form
	double *tempPar = new double[m_iNParameters];
	for(i = 0; i < m_iNParameters; i++)
	{
		tempPar[i] = 0.0;
		for(int j = 0; j < m_iNParameters; j++)
		{
			tempPar[i] += m_pdHessian[i][j]*pParDelta[j];
		}
	}
	for(i = 0; i < m_iNParameters; i++)
	{
		quadratic += pParDelta[i]*tempPar[i];
	}
	delete [] tempPar;
	delete [] pRandVec;

	return quadratic;
}


bool CStochasticSensitivityAnalysis::AcceptMove(double quadratic, double deltaCost)
{
	double arg = (-1.0*deltaCost)/m_dTemperature;

	if(arg >= 0.0)
	{
		// accept the move
		return true;
	}
	else
	{
		double p = m_pRNG->uniform();
		if(p < exp(arg))
		{
			return true;
		}
		else
		{
			return false;
		}
	}
}

void CStochasticSensitivityAnalysis::AddParametersToEnsemble(double *pParSave, double *resid, double quadratic, double cost)
{
	m_vpEnsemble.push_back(new CEnsembleMember(m_iNParameters,pParSave,m_iNResiduals,resid,quadratic,cost));
}

void CStochasticSensitivityAnalysis::ConstructTrialMove(double *pParTrial, double *pParCurrent, double *pParDelta)
{
	m_pFilter->ForwardTransformation(pParCurrent,m_iNParameters);
	for(int j = 0; j < m_iNParameters; j++)
	{
		pParTrial[j] = pParCurrent[j] + pParDelta[j];
	}
	m_pFilter->BackwardTransformation(pParCurrent,m_iNParameters);
	m_pFilter->BackwardTransformation(pParTrial,m_iNParameters);
}

