// HessianMonteCarloQuench.cpp: implementation of the CHessianMonteCarloQuench class.
//
//////////////////////////////////////////////////////////////////////

#include "HessianMonteCarloQuench.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CHessianMonteCarloQuench::CHessianMonteCarloQuench(int nIterations,int seed,CParameterFilter *pFilter)
{
	m_iRNGSeed = seed;
	m_pFilter = pFilter;
	m_iNIterations = nIterations;
	m_pRNG = new Rand(seed);
	m_dControlParameter = 1.0;
}

CHessianMonteCarloQuench::~CHessianMonteCarloQuench()
{
	delete m_pRNG;
}

double CHessianMonteCarloQuench::Minimize(double *parameters, Minimizable *minimizable)
{
	// set to false if Hessian does not need to be recomputed
	bool computeFlag = true;
	// downcast the minimizable, and throw an exception if it is not an NLLSMinimizable 
	// (LevenbergMarquardt should only be used on a sum of squares objective function)
	NLLSMinimizable *nlls = dynamic_cast<NLLSMinimizable *>(minimizable);
	if(0 == nlls){throw std::runtime_error("Minimizable is not a sum of squares!");}

	// get the number of parameters and allocate space for the Hessian
	nParameters = nlls->GetNParameters();
	m_iNResiduals = nlls->GetNResiduals();
	
	// Since a computation of the Hessian requires nParameters cost computations, we 
	// perform N moves between every computation to balance the computing cost
	m_iNUpdates = nParameters; 

	m_pdSigmaInverse = new double *[nParameters];
	m_pdCholeskyFactor = new double *[nParameters];
	m_pdSigmaInverse[0] = new double[nParameters*nParameters];
	m_pdCholeskyFactor[0] = new double[nParameters*nParameters];
	for(int nPar = 1; nPar < nParameters; nPar++)
	{
		m_pdSigmaInverse[nPar] = &(m_pdSigmaInverse[0][nPar*nParameters]);
		m_pdCholeskyFactor[nPar] = &(m_pdCholeskyFactor[0][nPar*nParameters]);
	}

	// allocate space for current and trial parameters
	double *parCurrent = new double[nParameters];
	double *parTrial = new double[nParameters];

	for(int i = 0; i < nParameters; i++)
	{
		parCurrent[i] = parameters[i];
		parTrial[i] = parameters[i];
	}

	// obtain the current cost
	double CurrentCost = nlls->ObjectiveFunction(parCurrent);

	for(int nIt = 0; nIt < m_iNIterations; i++)
	{
		// compute sigma inverse, then invert sigma and decompose (if sigma is not recomputed none 
		// of these steps need to be repeated)
		if(computeFlag)
		{
			ComputeDerivativeInformation(parCurrent,nlls);
			ComputeCholeskyFactor();
		}

		cin.get();
		
		// generate gaussian dist. random numbers using Cholesky compsition

		// move and accept if cost is lowered
	}

	return m_dBestCost;
}

void CHessianMonteCarloQuench::ComputeDerivativeInformation(double *parameters,NLLSMinimizable *nlls)
{
	double transparameter = 0.0;
	double saveparameter = 0.0;
	//double gradNorm = 0.0;
	// allocate
	double *fxPlus = new double[m_iNResiduals];
	double *fx = new double[m_iNResiduals];
	double **jacobian = new double *[m_iNResiduals];
	jacobian[0] = new double[m_iNResiduals*nParameters];
	for(int nR = 1; nR < m_iNResiduals; nR++)
	{
		jacobian[nR] = &(jacobian[0][nR*nParameters]);
	}

	// XXX debug
	double **jTrans = new double *[nParameters];
	jTrans[0] = new double[nParameters*m_iNResiduals];
	for(int nP = 1; nP < nParameters; nP++)
	{
		jTrans[nP] = &(jTrans[0][nP*m_iNResiduals]);
	}

	// get residuals at starting point
	double chiSq = nlls->ObjectiveFunction(parameters);
	for(int k = 0; k < m_iNResiduals; k++)
	{
		fx[k] = (nlls->GetResiduals())[k];
	}

	// compute Jacobian
	for(int i = 0; i < nParameters; i++)
	{
		// transform the parameter and save its old value
		saveparameter = parameters[i];
		transparameter = m_pFilter->Operator(parameters[i]);

		// if the parameter happens to equal 0.0, the stepsize will then be 
		// zero and give NaN upon division, so we set it to 1.0e-08
		double h = 1.0e-08;
		if(transparameter > 0.0)
		{
			h = pow(DBL_EPSILON,0.333333)*transparameter;
			//h = DBL_EPSILON*transparameter;
		}
				
		// forward step
		parameters[i] = m_pFilter->OperatorInverse(transparameter + h);
		double chiSqPlus = nlls->ObjectiveFunction(parameters);
		for(int k = 0; k < m_iNResiduals; k++)
		{
			fxPlus[k] = (nlls->GetResiduals())[k];
		}

		// compute scaled gradient element
		//m_pdBeta[i] = -0.5*((chiSqPlus - chiSq)/h);
		//gradNorm += m_pdBeta[i]*m_pdBeta[i];
		
		parameters[i] = saveparameter;

		// fill in jacobian
		for(int j = 0; j < m_iNResiduals; j++)
		{
			double delta = (fxPlus[j] - fx[j])/h;
			//jacobian[i][j] = -1*delta;
			jacobian[j][i] = -1*delta;
		}
	}
	
	// might want to use this to break out of the Monte Carlo?
	// compute gradient norm and check it
	//gradNorm = sqrt(gradNorm);
	//if(gradNorm < m_dGradTol) {m_bBreakFlag = true;}

	// compute the scaled hessian (it is symmetric)
	for(i = 0; i < nParameters; i++)
	{
		for(int k = i; k < nParameters; k++)
		{
			m_pdSigmaInverse[i][k] = 0.0;
			for(int j = 0; j < m_iNResiduals; j++)
			{
				m_pdSigmaInverse[i][k] += jacobian[j][i]*jacobian[j][k];
			}
			m_pdSigmaInverse[k][i] = m_pdSigmaInverse[i][k];
		}
	}

	// XXX debug
	// A check on the hessian
	for(int row = 0; row < m_iNResiduals; row++)
	{
		for(int col = 0; col < nParameters; col++)
		{
			jTrans[col][row] = jacobian[row][col];
		}
	}
	
	for(int l = 0; l < nParameters; l++)
	{
		for(int m = 0; m < nParameters; m++)
		{
			double testElem = 0.0;
			for(int n = 0; n < m_iNResiduals; n++)
			{
				testElem += jTrans[l][n]*jacobian[n][m]; 
			}
			cout << testElem << " " << m_pdSigmaInverse[l][m] << endl;
		}
	}

	delete [] fxPlus;
	delete [] fx;
	delete [] jacobian[0];
	delete [] jacobian;

	// XXX debug
	delete [] jTrans[0];
	delete [] jTrans;

	cout << "Hessian calculated . . ." << endl;

	return;

}

void CHessianMonteCarloQuench::ComputeCholeskyFactor()
{
// we're giving the LAPACK routine the upper triangle (though it doesn't matter)
	char UPLO = 'U';
	char TRANS = 'N';
	// can check for correct factorization
	int INFO=0; 
	// order of A
	int N = nParameters;
	int M = nParameters;
	// need to be >= max(1,N);
	int LDA = nParameters;
	int LDB = nParameters;
	// only one right hand side
	int NRHS  = 1;
	
	// in order to invert the matrix, we first have to LU decompose it
	double *A = new double[nParameters*nParameters];
	int *IPIV = new int[nParameters];
	double *WORK = new double[nParameters];
	int counter = 0;
	for(int j = 0; j < nParameters; j++)
	{
		for(int i = 0; i < nParameters; i++)
		{
			A[counter] = m_pdSigmaInverse[i][j];
			cout << A[counter] << endl;
			counter++;
		}
	}
	// LU decompose
	DGETRF(&M,&N,A,&LDA,IPIV,&INFO);
	if(INFO != 0)
	{
		cout << "LU decomp : INFO = " << INFO << endl;
		throw std::runtime_error("LU decomposition failed");
	}

	// now invert the LU decomposed matrix
	DGETRI(&N,A,&LDA,IPIV,WORK,&N,&INFO);
	if(INFO != 0)
	{
		cout << "Inverse : INFO = " << INFO << endl;
		throw std::runtime_error("Inversion of LU decomposed matrix failed");
	}

	/*
	// XXX debug
	for(int l = 0; l < nParameters*nParameters; l++)
	{
		cout << A[l] << endl;
	}
	*/

	// now cholesky decompose the inverted matrix
	double *B = new double[(nParameters*(nParameters + 1))/2];
	int columnCounter = 0;
	int Bcounter = 0;
	for(int i = 0; i < nParameters*nParameters; i+= nParameters)
	{
		for(int j = 0; j <= columnCounter; j++)
		{
			B[Bcounter] = A[i+j];
			//cout << B[Bcounter] << endl;
			Bcounter++;
		}
		columnCounter++;
	}
	// Cholesky decompose
	DPOTRF(&UPLO,&N,B,&LDA,&INFO);
	if(INFO != 0)
	{
		cout << "Cholesky factorization : INFO = " << INFO << endl;
		throw std::runtime_error("Cholesky factorization failed");
	}
	
	/*
	// copy the matrix square root into our own matrix
	for(int i = 0; i < nParameters; i++)
	{
		deltaP[i] = B[i];
	}
	*/	
	
	delete [] A;
	delete [] B;
	delete [] IPIV;
	delete [] WORK;
}