// SingleNetworkRateConstantMinimizable.cpp: implementation of the CSingleNetworkRateConstantMinimizable class.
//
//////////////////////////////////////////////////////////////////////

#include "SingleNetworkRateConstantMinimizable.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CSingleNetworkRateConstantMinimizable::CSingleNetworkRateConstantMinimizable()
{
	m_dTimeSeriesWeight = 1.0;
	m_dRateConstantsWeight = 0.0;
	m_bLogsInObjectiveFunction = false;
}

CSingleNetworkRateConstantMinimizable::~CSingleNetworkRateConstantMinimizable()
{
	delete [] m_pdInitialNetworkData;
	delete [] m_pdCurrentNetworkData;
	for(int i = 0; i < m_pConversionFactors.size(); i++)
	{
		delete m_pConversionFactors[i];
	}
	m_pReactionMover->Detach(m_pCellObserver);
	delete m_pCellObserver;
}

CSingleNetworkRateConstantMinimizable::CSingleNetworkRateConstantMinimizable(Experiment *pExperiment, CReactionMover *pReactionMover, bool logsInObjectiveFunction, double timeSeriesWeight, double rateConstantsWeight, int nCells)
{
	Initialize(pExperiment, pReactionMover, logsInObjectiveFunction, timeSeriesWeight, rateConstantsWeight, nCells);
}

CSingleNetworkRateConstantMinimizable::Initialize(Experiment *pExperiment, CReactionMover *pReactionMover, bool logsInObjectiveFunction, double timeSeriesWeight, double rateConstantsWeight, int nCells)
{
	m_pExperiment = pExperiment;
	m_pReactionMover = pReactionMover;
	m_iNCells = nCells;
	m_bLogsInObjectiveFunction = logsInObjectiveFunction;
	m_dTimeSeriesWeight = timeSeriesWeight;
	m_dRateConstantsWeight = rateConstantsWeight;

	// Read in the data
	m_pExperiment->ReadData();
	// Initialize variables dependent on the Experiment class
	int nChem = m_pExperiment->GetReactionNetwork()->GetNumberOfChemicals();
	int nRateConstants = m_pExperiment->GetReactionNetwork()->GetNumberOfRateConstants();
	int intTime = ((int) m_pExperiment->GetChemicalTimeSeriesData()->GetLatestDataTime()) + 2;
	m_pCellObserver = new CellAverageObserver(nChem,intTime,nCells);

	// attach the observer
	m_pReactionMover->Attach(cellObserver);

	m_pdInitialNetworkData = new double[nRateConstants + nChem];
	m_pdCurrentNetworkData = new double[nRateConstants + nChem];

	// load up the vector of conversion factors with dummies assuming all data is 
	// expressed in terms of absolute numbers of molecules; if we don't have data 
	// we don't need conversion factors
	for(int l = 0; l < nChem; l++)
	{
		if((experiment->GetChemicalTimeSeriesData()->GetTimeSeries(l)).size() > 0 )
		{
			m_pConversionFactors.push_back(new CConversionFactor(true,false,1.0));
		}
		else
		{
			m_pConversionFactors.push_back(new CConversionFactor(false));
		}
	}

	// store beginning network data - this allows us to restore the 
	// initial state of the network to recover from the beginning
	for(int i = 0; i < nRateConstants; i++)
	{
		m_pdInitialNetworkData[i] = m_pExperiment->GetReactionNetwork()->GetRateConstant(i)->GetInitialValue();
		m_pdCurrentNetworkData[i] = m_pdInitialNetworkData[i];
	}
	for(i = nRateConstants; i < nRateConstants + nChem; i++)
	{
		m_pdInitialNetworkData[i] = m_pExperiment->GetReactionNetwork()->GetChemical(i-nRateConstants)->GetInitialAmount();
		m_pdCurrentNetworkData[i] = m_pdInitialNetworkData[i];
	}

	int localResidualsSize = 0;
	if(m_dTimeSeriesWeight > 0.0) {localResidualsSize += this->experiment->GetChemicalTimeSeriesData()->GetNDataPoints();}
	if(m_dRateConstantsWeight > 0.0) {localResidualsSize += nRateConstants;}

	Allocate(localResidualsSize);
}

double CSingleNetworkRateConstantMinimizable::ComputeResiduals(double *parameters)
{
	double objFuncVal = 0.0;
	int nRateConstants = m_pExperiment->GetReactionNetwork()->GetNumberOfRateConstants();
	int nChem = m_pExperiment->GetReactionNetwork()->GetNumberOfChemicals();
	
	// put the parameters (rate constants) into the network	
	for(int i = 0; i < nRateConstants; i++)
	{
		m_pExperiment->GetReactionNetwork()->GetRateConstant(i)->SetRateConstant(parameters[i]);
		m_pdCurrentNetworkData[i] = parameters[i];
	}
	m_pExperiment->GetReactionNetwork()->ChemicalReset();
	
	// attach the observer and zero him
	cellObserver->ZeroConcentrations();

	for(int cellCounter = 1; cellCounter <= nCells; cellCounter++)
	{
		m_pReactionMover->Move(0.0,experiment->GetChemicalTimeSeriesData()->GetLatestDataTime(),experiment->GetReactionNetwork());
	}

	// compute the scaling factors; successive computation of the cost is somewhat of a
	// repeat of this operation, so combining the two calculations in the future may 
	// speed things up
	for(int chemCount = 0; chemCount < nChem; chemCount++)
	{
		double num = 0.0;
		double den = 0.0;
		ChemicalTimeSeriesData *cTSD = m_pExperiment->GetChemicalTimeSeriesData();
		// do we need a scaling factor AND is it allowed to vary?
		if(m_pConversionFactors[chemCount]->IsFactorNeeded())
		{
			if(m_pConversionFactors[chemCount]->IsFactorFixed())
			{
				num = m_pConversionFactors[chemCount]->GetFactorValue();
				den = 1.0;
			}
			else
			{
				int dataSize = (cTSD->GetTimeSeries(chemCount)).size();
				// loop over the number of data points and compute
				for(int nData = 0; nData < dataSize; nData++)
				{
					double time = cTSD->GetTimeSeries(chemCount)[nData]->GetTime();
					double data = cTSD->GetTimeSeries(chemCount)[nData]->GetDatum();
					double error = cTSD->GetTimeSeries(chemCount)[nData]->GetError();
					double sim = (m_pCellObserver->GetAverageConcentration())[chemCount][int(time)];
					num += ((2*sim)/(error*error))*data;
					den += ((2*sim)/(error*error))*sim;
				}
			}
		}
		else
		{
			num = 1.0;
			den = 1.0;
		}

		double factor = num/den;
		m_pConversionFactors[chemCount]->SetFactorValue(factor);
	}
		
	// compute the cost
	int nextAvailableIndex = 0;
	if(m_dTimeSeriesWeight > 0.0)
	{
		// load the time series residuals in the master list
		for(int chemIt = 0; chemIt < (m_pExperiment->GetChemicalTimeSeriesData()->GetTimeVector()).size(); chemIt++)
		{	
			// get the keys
			double targetTime = (m_pExperiment->GetChemicalTimeSeriesData()->GetTimeVector())[chemIt]->GetTime();
			int targetChem = (m_pExperiment->GetChemicalTimeSeriesData()->GetTimeVector())[chemIt]->GetChemNumber();
			int vectorIndex = (m_pExperiment->GetChemicalTimeSeriesData()->GetTimeVector())[chemIt]->GetVectorIndex();
			// use the keys
			double data = (m_pExperiment->GetChemicalTimeSeriesData()->GetTimeSeries(targetChem))[vectorIndex]->GetDatum();
			double sigma = (m_pExperiment->GetChemicalTimeSeriesData()->GetTimeSeries(targetChem))[vectorIndex]->GetError();
			double sim = (m_pCellObserver->GetAverageConcentration())[targetChem][int(targetTime)];
			// rescale sim if there is a conversion factor
			if(m_pConversionFactors[targetChem]->IsFactorNeeded())
			{
				sim = sim*(m_pConversionFactors[targetChem]->GetFactorValue());
			}
			// compute quantities of interest
			residuals[nextAvailableIndex] = (data - sim)/sigma;
			objFuncVal += residuals[nextAvailableIndex]*residuals[nextAvailableIndex]*m_dTimeSeriesWeight;
			nextAvailableIndex++;
		}
	}
	if(m_dRateConstantsWeight > 0.0)
	{
		// load in the rate constants residuals in the master list
		if(m_bLogsInObjectiveFunction)
		{
			for(int i = 0; i < nRateConstants; i++)
			{
				double sigma = m_pExperiment->GetReactionNetwork()->GetRateConstant(i)->GetErrorInInitialValue();
				double sigmalog = sigma/m_pdInitialNetworkData[i];
				residuals[nextAvailableIndex] = log(m_pdInitialNetworkData[i]/m_pdCurrentNetworkData[i])/sigmalog;
				objFuncVal += residuals[nextAvailableIndex]*residuals[nextAvailableIndex]*m_dRateConstantsWeight;
				nextAvailableIndex++;
			}
		}
		else
		{
			for(int i = 0; i < nRateConstants; i++)
			{
				double sigma = m_pExperiment->GetReactionNetwork()->GetRateConstant(i)->GetErrorInInitialValue();
				residuals[nextAvailableIndex] = (m_pdInitialNetworkData[i] - m_pdCurrentNetworkData[i])/sigma;
				objFuncVal += residuals[nextAvailableIndex]*residuals[nextAvailableIndex]*m_dRateConstantsWeight;
				nextAvailableIndex++;
			}
		}
	}
	
	// notify observers
	Notify();
	
	return objFuncVal;
}

int CSingleNetworkRateConstantMinimizable::GetNParameters()
{
	return m_pExperiment->GetReactionNetwork()->GetNumberOfRateConstants();
}

double CSingleNetworkRateConstantMinimizable::GetParameter(int parIndex)
{
	int nP = GetNParameters();
	
	if(parIndex > nP)
	{
		throw std::runtime_error("Parameter index is off parameter list.");
	}
	else
	{
		return m_pdCurrentNetworkData[parIndex];
	}

	return 0.0;

}